package cn.org.bjca.trust.java.imserver.im;


import cn.org.bjca.trust.java.imserver.common.SpringUtilsAuTo;
import cn.org.bjca.trust.java.imserver.common.json.GsonImplHelp;
import cn.org.bjca.trust.java.imserver.entitys.GroupEntity;
import cn.org.bjca.trust.java.imserver.enums.PacketType;
import cn.org.bjca.trust.java.imserver.im.bean.PacketMessage;
import cn.org.bjca.trust.java.imserver.im.msg.ConnectMessage;
import cn.org.bjca.trust.java.imserver.im.msg.ConnectedMessage;
import cn.org.bjca.trust.java.imserver.im.msg.ConversationMessage;
import cn.org.bjca.trust.java.imserver.im.msg.msg.SZYXMessage;
import cn.org.bjca.trust.java.imserver.im.msg.msg.SendAckMessage;
import cn.org.bjca.trust.java.imserver.repository.*;
import cn.org.bjca.trust.java.imserver.request.RequestManager;
import org.eclipse.paho.client.mqttv3.*;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;

import java.nio.charset.StandardCharsets;
import java.util.List;

public class ImManager {

    private final UserRepository userRepository = SpringUtilsAuTo.getBean(UserRepository.class);
    private final MessageRepository messageRepository = SpringUtilsAuTo.getBean(MessageRepository.class);
    private final ConversationRepository conversationRepository = SpringUtilsAuTo.getBean(ConversationRepository.class);
    private final GroupRepository groupRepository = SpringUtilsAuTo.getBean(GroupRepository.class);
    private final ConnectRepository connectRepository = SpringUtilsAuTo.getBean(ConnectRepository.class);

    public static ImManager getInstance() {
        return ImManagerHolder.instance;
    }

    private static final class ImManagerHolder {
        static final ImManager instance = new ImManager();
    }

    private MqttClient mqttClient;
    private final MqttConnectOptions connectOptions;

    public ImManager() {
        connectOptions = new MqttConnectOptions();
        connectOptions.setCleanSession(false);
        connectOptions.setUserName("server");
        connectOptions.setPassword("server".toCharArray());
        connectOptions.setConnectionTimeout(30);
        connectOptions.setKeepAliveInterval(10);
        connectOptions.setAutomaticReconnect(true);


        try {
            mqttClient = new MqttClient("tcp://114.115.203.60:18883",
                    "server" + System.currentTimeMillis(), new MemoryPersistence());
            mqttClient.setCallback(new MqttCallbackExtended() {
                @Override
                public void connectComplete(boolean reconnect, String serverURI) {
                    System.out.println("======>connectComplete:reconnect?" + reconnect + "::" + serverURI);
                    if (reconnect) return;
                    try {
                        mqttClient.subscribe("server" + PacketType.CONNECT, 2);
                        mqttClient.subscribe("server" + PacketType.SEND, 2);
                        mqttClient.subscribe("server" + PacketType.PING, 2);
                        mqttClient.subscribe("server" + PacketType.REQUEST, 2);
                    } catch (MqttException ignored) {
                    }
                }

                @Override
                public void connectionLost(Throwable cause) {
                    System.out.println("======>connectionLost" + cause.getMessage());
                }

                @Override
                public void messageArrived(String topic, MqttMessage message) throws Exception {
                    String msg = new String(message.getPayload());
                    System.out.println("======>messageArrived:\n" + topic + "::" + msg);
                    if (topic.equals("server" + PacketType.CONNECT)) {
                        sendConnectedMsg(GsonImplHelp.get().toObject(msg, ConnectMessage.class));
                    } else if (topic.equals("server" + PacketType.SEND)) {
                        SZYXMessage szyxMessage = GsonImplHelp.get().toObject(msg, SZYXMessage.class);
                        szyxMessage.setStatus(0);
                        saveConversation(szyxMessage);

                        SZYXMessage message1 = szyxMessage.copy();
                        message1.setArrive(true);
                        message1.setStatus(1);
                        if (szyxMessage.isGroup()) {
                            for (GroupEntity groupEntity : groupRepository.findAllByAppIdAndGroupIdAndUserIdNot(message1.getAppId(), message1.getGroupID(), message1.getFromClientId())) {
                                SZYXMessage message2 = message1.copy();
                                message2.setUserName(groupEntity.getUserId());
                                saveConversation(message2);
                                sendPacketMessage(PacketType.SEND, message2);
                            }
                        } else {
                            message1.setUserName(message1.getToClientId());
                            saveConversation(message1);
                            sendPacketMessage(PacketType.SEND, message1);
                        }
                        // TODO: 这里应该有计时器，判断转发消息是否送达，如果没送达则ack返回失败状态
                        sendAckMsg(szyxMessage);

                    }// 数据请求
                    else if (topic.equals("server" + PacketType.REQUEST)) {
                        sendPacketMessage(PacketType.RESULT, RequestManager.getInstance().request(msg));
                    }

                }

                @Override
                public void deliveryComplete(IMqttDeliveryToken token) {
                }

            });
        } catch (MqttException e) {
            System.out.println("======>MqttException" + e.getMessage());
        }

    }

    private void saveConversation(SZYXMessage szyxMessage) {
        try {
            List<ConversationMessage> l = conversationRepository.getConversation(szyxMessage.getAppId(), szyxMessage.getUserName(),
                    szyxMessage.isArrive() ? szyxMessage.getFromClientId() : szyxMessage.getToClientId(), szyxMessage.getGroupID());
            ConversationMessage conversation;
            if (l.size() == 0) {
                conversation = new ConversationMessage();
            } else {
                conversation = l.get(0);
            }
//            ConversationMessage conversation = new ConversationMessage();
            conversation.setGroup(szyxMessage.isGroup());
            conversation.setSender(!szyxMessage.isArrive());
            conversation.setMessage(szyxMessage);

            conversation.setAppId(szyxMessage.getAppId());
            conversation.setUserName(szyxMessage.getUserName());
            conversation.setOsType(szyxMessage.getOsType());
            conversation.setToken(szyxMessage.getToken());

            if (szyxMessage.isGroup()) {
                conversation.setGroupEntity(groupRepository.findFirstByAppIdAndGroupId(szyxMessage.getAppId(), szyxMessage.getGroupID()));
            } else {
                conversation.setUserId(szyxMessage.isArrive() ? szyxMessage.getFromClientId() : szyxMessage.getToClientId());
            }

            conversationRepository.save(conversation);
            sendPacketMessage(PacketType.CONVERSATION, conversation);
        } catch (Exception e) {
            System.out.println("--------------" + e.getMessage());
        }
    }

    public void connect() {
        new Thread(() -> {
            try {
                if (null != mqttClient) {
                    if (mqttClient.isConnected()) {
                        mqttClient.disconnect();
                    }
                    mqttClient.connect(connectOptions);
                }
            } catch (MqttException e) {
                System.out.println("======>connect" + e.getMessage());
            }
        }
        ).start();
    }


    private void sendConnectedMsg(ConnectMessage connectMessage) {
        try {
            connectRepository.save(connectMessage.copy());
        } catch (Exception ignored) {

        }
        ConnectedMessage message = new ConnectedMessage();
        message.setStatus(200);
        message.setAppId(connectMessage.getAppId());
        message.setUserName(connectMessage.getUserName());
        sendPacketMessage(PacketType.CONNECTED, message);
    }

    private void sendAckMsg(SZYXMessage szyxMessage) {
        SendAckMessage message = new SendAckMessage();
        message.setAckId(szyxMessage.getMsgId());
        message.setAckStatus(0);
        message.setAppId(szyxMessage.getAppId());
        message.setUserName(szyxMessage.getUserName());
        sendPacketMessage(PacketType.SEND_ACK, message);
    }

    private void sendPacketMessage(PacketType packetType, PacketMessage message) {
        if (null == connectOptions) {
            return;
        }
        String msg = GsonImplHelp.get().toJson(message);
        try {
            String topic = message.getAppId() + "/" + packetType + message.getUserName();
            mqttClient.publish(topic, msg.getBytes(StandardCharsets.UTF_8), 2, false);

            System.out.println("======>发送消息：\n" + topic + "\n" + msg);
        } catch (MqttException ignored) {

        }

    }


}
